module FoldsMonoidsFoldables where

import Data.Foldable
import qualified Data.List.NonEmpty as NE
import Data.Monoid

sum' :: Num a => [a] -> a
sum' [] = 0
sum' (x:xs) = x + sum' xs

product' :: Num a => [a] -> a
product' [] = 1
product' (x:xs) = x * product' xs

concat' :: [[a]] -> [a]
concat' [] = []
concat' (x:xs) = x ++ concat' xs

foldrList :: (a -> b -> b) -> b -> [a] -> b
foldrList f init [] = init
foldrList f init (x:xs) = f x $ foldrList f init xs

sum'' = foldrList (+) 0
product'' = foldrList (*) 1
concat'' = foldrList (++) []
length'' = foldrList (\_ acc -> acc+1) 0

ls = map show [1..5]
foldrEx = foldr (\x acc -> mconcat ["(",x,"+",acc,")"] ) "0" ls

foldlList :: (b -> a -> b) -> b -> [a] -> b
foldlList f init [] = init
foldlList f init (x:xs) = foldlList f (f init x) xs

foldlList' :: (b -> a -> b) -> b -> [a] -> b
foldlList' f init [] = init
foldlList' f init (x:xs) = acc `seq` foldlList' f acc xs
                            where acc = f init x

-- foldr1 (-) [1,2,3]
-- (1-(2-3))
-- foldl1 (-) [1,2,3]
-- ((1-2)-3)

-- scanr
-- scanl

unfoldrList :: (b->Maybe(a,b)) -> b -> [a]
unfoldrList f b = case f b of
                   Nothing -> []
                   Just (x, newB) -> x : unfoldrList f newB

unfoldrExample =
        unfoldrList (\x -> if x == 0 then Nothing else Just (x, x-1)) 10

{-- Semigroup --}
-- <>
neList1 :: NE.NonEmpty Int
neList1 = 1 NE.:|[2,3,4]

neList2 :: NE.NonEmpty Int
neList2 = 5 NE.:|[6,7,8]

neList3 = neList1 <> neList2

{-- Monoid --}
data List a = Nil | Cons a (List a) deriving Show

concatList :: List a -> List a -> List a
concatList Nil xs = xs
concatList (Cons x xs) ys = Cons x (concatList xs ys)

list1 = Cons 1 (Cons 2 Nil)
list2 = Cons 3 (Cons 4 (Cons 5 Nil))

instance Semigroup (List a) where
    (<>) = concatList

instance Monoid (List a) where
    mempty = Nil

mconcat' :: Monoid a => [a] -> a
mconcat' = foldr (<>) mempty

newtype Sum' a = Sum' a deriving Show

instance Num a => Semigroup (Sum' a) where
  (Sum' x) <> (Sum' y) = Sum' $ x+y

instance Num a => Monoid (Sum' a) where
  mempty = Sum' 0

-- mconcat $ map Sum' [1,2,3]

 -- (||) mempty = False : Any
 -- (&&) mempty = True  : All
 -- getAny $ mconcat $ map Any [True, False, True]    True

-- Dual getDual . mconcat . map Dual $ ["abc", "def"]

-- Endo  ( mconcat $  map Endo [(+2), (*2), (^3)]) `appEndo` 1

{-- Foldable --}
data Tree a = Empty | Tree a (Tree a) (Tree a) deriving Show
-- :k Either Int
-- Either Int :: * -> *

instance Foldable Tree where
  foldMap f Empty = mempty
  foldMap f (Tree v l r) =
           (foldMap f l) <> f v <> (foldMap f r)

t = Tree 2 Empty (Tree 3 (Tree 4 Empty Empty) (Tree 5 Empty Empty))

-- getSum . foldMap (Sum) $ t
-- 14
-- getProduct . foldMap (Product) $ t
-- 120
